# -*- coding: utf-8 -*-
"""
---------------------------------------------
Created on 2025/3/5 21:40
@author: ZhangYundi
@email: yundi.xxii@outlook.com
---------------------------------------------
"""
from __future__ import annotations

from .qdf import QDF
from .lazy import LQDF
from .expr import Expr
from typing import TYPE_CHECKING
from pathlib import Path
from ..dataset import scan

if TYPE_CHECKING:
    import polars as pl



def from_polars(df: pl.DataFrame | pl.LazyFrame | Path | str, index: tuple[str] = ("date", "time", "asset"), align: bool = False, ) -> QDF:
    """polars dataframe 转为 表达式数据库"""
    if isinstance(df, (Path, str)):
        df = scan(df)
    return QDF(df, index, align,)

# def to_lazy(df: pl.DataFrame | pl.LazyFrame | Path | str, index: tuple[str] = ("date", "time", "asset"), align: bool = False, ) -> LQDF:
#     """polars dataframe/lazy frame/table path 转为 表达式数据库"""
#     if isinstance(df, (Path, str)):
#         df = scan(df)
#     return LQDF(df, index, align,)




# -*- coding: utf-8 -*-
"""
---------------------------------------------
Created on 2025/3/5 21:40
@author: ZhangYundi
@email: yundi.xxii@outlook.com
---------------------------------------------
"""
from __future__ import annotations

import importlib.util
import sys
from functools import lru_cache
from functools import partial
from pathlib import Path
import inspect

import polars as pl
from polars import selectors as cs
import logair

from .errors import CalculateError, CompileError, PolarsError, FailError
from .expr import Expr
from .expr import build_dependency_graph, topological_sort

# 动态加载模块
module_name = "udf"
module_path = Path(__file__).parent / "udf" / "__init__.py"
spec = importlib.util.spec_from_file_location(module_name, module_path)
module = importlib.util.module_from_spec(spec)
sys.modules[module_name] = module
spec.loader.exec_module(module)

logger = logair.get_logger(__name__)

class LQDF:

    def __init__(self,
                 data: pl.LazyFrame | pl.DataFrame,
                 index: tuple[str] = ("date", "time", "asset"),
                 align: bool = True, ):
        assert isinstance(data, (pl.LazyFrame, pl.DataFrame)), "data must be a polars DataFrame or LazyFrame"
        self.data: pl.LazyFrame = data.lazy().cast({pl.Decimal: pl.Float64}).fill_nan(None).drop_nulls(subset=index)
        self._index = self.data.select(index).collect()
        self.dims = [self._index[name].n_unique() for name in index]
        if align:
            lev_vals: list[pl.LazyFrame] = [self._index.select(name).unique().lazy() for name in index]
            full_index = lev_vals[0]
            for lev_val in lev_vals[1:]:
                full_index = full_index.join(lev_val, how="cross")
            self.data = full_index.join(self.data, on=index, how='left')
        self.data = self.data.sort(index)
        self.failed = list()
        self._cols = set(self.data.collect_schema().names())
        self._expr_cache = dict()  # type: dict[Expr, str]
        self._cur_expr_cache = dict()
        self._fn_map = dict()

    def __str__(self):
        return self.data.__str__()

    def __repr__(self):
        return self.data.__str__()

    def register_udf(self, func: callable, name: str = None):
        name = name if name is not None else func.__name__
        setattr(module, name, func)

    def _to_pl_expr(self, e: Expr) -> pl.Expr:
        """递归编译"""
        alias = e.alias
        if alias in self._cols:
            return pl.col(alias)
        elif e in self._expr_cache:
            return pl.col(self._expr_cache[e]).alias(alias)
        elif e in self._cur_expr_cache:
            return pl.col(self._cur_expr_cache[e]).alias(alias)
        else:
            func = self._fn_map.get(e.fn_name)
            if func is None:
                func = getattr(module, e.fn_name)
                _params = sorted(list(inspect.signature(func).parameters.keys()))
                if "dims" in _params:
                    func = partial(func, dims=self.dims)
                self._fn_map[e.fn_name] = func

            args = list()
            kwargs = dict()
            for arg in e.args:
                if isinstance(arg, Expr):
                    if arg in self._expr_cache:
                        args.append(pl.col(self._expr_cache[arg]))
                    elif arg in self._cur_expr_cache:
                        args.append(pl.col(self._cur_expr_cache[arg]))
                    # else:
                        # 拓扑解析依赖结构出错
                elif isinstance(arg, dict):
                    kwargs.update(arg)
                elif isinstance(arg, str):
                    if arg.lower() == "null":
                        args.append(None)
                    else:
                        args.append(pl.col(arg))
                else:
                    args.append(arg)  # or args.append(pl.lit(arg))
            try:
                expr_pl: pl.Expr = func(*args, **kwargs)
                self._cur_expr_cache[e] = alias
                return expr_pl.alias(alias)
            except Exception as error:
                raise CompileError(message=f"{e.fn_name}({', '.join([str(arg) for arg in args])})\n{error}") from error


    def sql(self, *exprs: str, ) -> pl.DataFrame:
        """
        表达式查询
        Parameters
        ----------
        exprs: str
            表达式，比如 "ts_mean(close, 5) as close_ma5"
        Returns
        -------
            polars.LazyFrame
        """
        self.failed = list()
        # exprs_select = list()
        exprs_parsed = list()
        self._cur_expr_cache = {}
        for expr in exprs:
            try:
                # compiled, alias = self._compile_expr(expr)
                e_parsed = Expr(expr)
                exprs_parsed.append(e_parsed)
            except Exception as e:
                self.failed.append(FailError(expr, e))
        if self.failed:
            logger.warning(f"sql failed num：{len(self.failed)}/{len(exprs)}: \n {self.failed}")
        graph, indegree, expr_map = build_dependency_graph(exprs_parsed, self._cols)
        lvls: list[list[Expr]] = topological_sort(graph, indegree, expr_map)
        for batch_exprs in lvls:
            batch_exprs = [self._to_pl_expr(e) for e in batch_exprs]
            self.data = self.data.with_columns(*batch_exprs).fill_nan(None)
        new_expr_cache = dict()
        try:
            current_cols = set(self.data.collect_schema().names())
            # 缓存整理：只保留当前表达式的缓存
            self._expr_cache.update(self._cur_expr_cache)
            for k, v in self._expr_cache.items():
                if v in current_cols:
                    new_expr_cache[k] = v
            self._expr_cache = new_expr_cache
            final_df = self.data.select(*self._index.columns, *[e.alias for e in exprs_parsed])
            return final_df
            # return final_df
        except Exception as e:
            # 缓存整理：只保留当前表达式的缓存
            for k, v in self._expr_cache.items():
                if v in self._cols:
                    new_expr_cache[k] = v
            self._expr_cache = new_expr_cache
            raise PolarsError(message=f"LazyFrame.collect() step error:\n{e}") from e



# Copyright (c) ZhangYundi.
# Licensed under the MIT License. 
# Created on 2025/11/10 13:43
# Description: 只有一张表单，没有分区的dataset特例, 所有数据都在一张表中

from __future__ import annotations

import sys
from collections.abc import Callable
from enum import Enum

import xcals
from functools import partial
import polars as pl
from datetime import datetime
import logair
import uuid
from .database import tb_path, scan

import ygo


class TableMode(Enum):

    F = "full" # 全量更新
    I = "increment" # 增量更新

class Table:

    def __init__(self,
                 fn: Callable[..., pl.DataFrame],
                 tb: str,
                 update_time: str,
                 mode: TableMode = TableMode.F):
        self.fn = fn
        self.tb = tb
        self.update_time = update_time
        self._data_dir = tb_path(self.tb)
        self.logger = logair.get_logger(__name__)
        self.verbose = False
        self.mode = mode

    def __call__(self, *args, **kwargs):
        fn = partial(self.fn, *args, **kwargs)
        table = Table(fn,
                      tb=self.tb,
                      update_time=self.update_time,
                      mode=self.mode)
        return table

    def _log(self, msg: str, lvl: str = "info"):
        """统一日志输出方法"""
        if self.verbose:
            getattr(self.logger, lvl)(f"{self.tb}: {msg}")

    def _do_job(self):
        """获取数据并且保存数据"""
        data = ygo.delay(self.fn)(this=self)()
        if data is None:
            self.logger.error(f"{self.tb}: No data.")
            return
        if data.is_empty():
            self.logger.warning(f"{self.tb}: No data.")
            return
        if self.mode == TableMode.I:
            time_uuid = uuid.uuid1()
            data_file = self._data_dir / f"{time_uuid}.parquet"
            data.write_parquet(data_file)
        elif self.mode == TableMode.F:
            data_file = self._data_dir / "0.parquet"
            data.write_parquet(data_file)
        else:
            self.logger.error(f"Invalid table mode: {self.mode}")


    def update(self, verbose: bool = False):
        """更新最新数据: 全量更新, 覆盖旧数据"""
        self.verbose = verbose
        if self._need_update(date=xcals.today()):
            self._log("Updating.", "info")
            self._do_job()

    def _need_update(self, date: str) -> bool:
        """是否需要更新"""
        existed = self._data_dir.exists()
        if not existed:
            self._data_dir.mkdir(parents=True, exist_ok=True)
            return True
        else:
            modified_time = self.modified_time
            if modified_time is not None:
                modified_datetime = modified_time.strftime("%Y-%m-%d %H:%M:%S")
                modified_d, modified_t = modified_datetime.split(" ")
                if self._updated(date, data_date=modified_d, data_time=modified_t):
                    return False
            return True

    def get_value(self, date: str, eager: bool = True) -> pl.DataFrame | pl.LazyFrame:
        """获取数据"""
        # self.update(verbose=True)
        if not date:
            date = xcals.today()
        self.verbose = True
        if self._need_update(date):
            self._log("Update first plz.", "warning")
            sys.exit()

        df = scan(self._data_dir)
        if eager:
            return df.collect()
        return df

    def _updated(self, date: str, data_date: str, data_time: str) -> bool:
        """判断是否已经更新数据"""
        recent_tradeday = xcals.get_recent_tradeday(date)
        prev_tradeday = xcals.shift_tradeday(recent_tradeday, -1)
        now = xcals.now()
        latest_update_date = recent_tradeday if now >= self.update_time else prev_tradeday
        return f"{data_date} {data_time}" >= f"{latest_update_date} {self.update_time}"

    @property
    def latest_file(self):
        if not self._data_dir.exists():
            return
        parquet_files = list(self._data_dir.glob("*.parquet"))
        if not parquet_files:
            return
        latest_file = max(parquet_files, key=lambda x: x.stat().st_mtime)
        return latest_file

    @property
    def modified_time(self):
        """获取文件修改时间"""
        latest_file = self.latest_file
        if latest_file is None:
            return
        mtime = self.latest_file.stat().st_mtime
        return datetime.fromtimestamp(mtime)



# Copyright (c) ZhangYundi.
# Licensed under the MIT License. 
# Created on 2025/12/11 10:41
# Description:

from collections import defaultdict, deque
from lidb.qdf.expr import Expr

def build_dependency_graph(exprs: list[Expr], avail_cols: set[str]) -> tuple[dict, dict, dict]:
    """
    构建表达式的依赖图
    Parameters
    ----------
    exprs: list[Expr]
    avail_cols: set[str]

    Returns
    -------
    tuple[dict, dict]
        - graph: {expr_alias: [依赖的 expr_alias]}
        - indegree: {expr_alias: indegree}
        - expr_map: {expr_alias: Expr}

    """

    graph = defaultdict(list)
    indegree = defaultdict(int)
    expr_map = {}

    def collect_deps(e: Expr):
        alias = e.alias
        expr_map[alias] = e
        for arg in e.args:
            if isinstance(arg, Expr):
                graph[arg.alias].append(alias)
                indegree[alias] += 1
                collect_deps(arg)
            elif isinstance(arg, str) and arg.lower() != "null":
                if arg not in avail_cols:
                    graph[arg].append(alias)
                    indegree[alias] += 1
    for expr in exprs:
        # parsed = parse_expr(expr) if isinstance(expr, str) else expr
        collect_deps(expr)

    for alias in expr_map:
        if alias not in indegree:
            indegree[alias] = 0

    return graph, indegree, expr_map

def topological_sort(graph: dict, indegree: dict, expr_map: dict) -> list[list[str]]:
    """
    返回按照层级划分的表达式执行顺序(每层内部没有依赖)
    """
    queue = deque([k for k in expr_map if indegree[k] == 0])
    levels = []
    while queue:
        level_size = len(queue)
        current_level = []
        for _ in range(level_size):
            node = queue.popleft()
            current_level.append(expr_map[node])
            for neighbor in graph[node]:
                indegree[neighbor] -= 1
                if indegree[neighbor] == 0:
                    queue.append(neighbor)
        levels.append(current_level)
    return levels

if __name__ == '__main__':
    exprs = ["high",
             "ma(close, 10) as ma_10",
             "ma(ma_10, 20) as ma_10_20",
             "std(osd_10, 30) as osd_10_30",
             "std(open, 10) as osd_10",
             "std(std(mean(close, 10), 30), 50) as c_ma10_sd30_sd50",
             "vol>0?amt/vol:null as vwap"]
    exprs = [Expr(e) for e in exprs]
    graph, indegree, expr_map = build_dependency_graph(exprs, avail_cols={"open", "high", "close", "vol", "amt"})
    print(graph)
    print(indegree)
    print(expr_map)
    lvl = topological_sort(graph, indegree, expr_map)
    print(lvl)



# Copyright (c) ZhangYundi.
# Licensed under the MIT License. 
# Created on 2025/12/11 14:52
# Description:
from omegaconf import OmegaConf

import lidb
import polars as pl
from lidb.qdf.lazy import LQDF
from lidb.qdf.lazy2 import LQDF as LQDF2
from lidb.qdf.qdf import QDF as LQDF3
import time
from hydra import main

pl.Config.set_tbl_cols(15)
pl.Config.set_tbl_width_chars(1500)
pl.Config.set_tbl_rows(30)

@main(version_base=None, config_path="./conf", config_name="svc-preds")
def run(cfg):
    quote_tb = lidb.tb_path("mc/stock_tick_cleaned")/ "freq=3s"/ "date=2025-05-06"
    quote = lidb.scan(quote_tb).filter(pl.col("time") <= "09:36:00")
    code_num = 600
    code_list = quote.select("asset").unique(maintain_order=True).collect()["asset"].to_list()
    codes = code_list[:code_num]
    quote = quote.filter(pl.col("asset").is_in(codes))
    start_t = time.time()
    db = LQDF3(quote, align=False, index=("date", "time", "asset"))
    print(f"LQDF elapsed: {(time.time()-start_t):.3f}s")
    # exprs = ["abs(c_ma20_ma30_sd30)", "ts_std(ts_mean(ts_mean(close, 20), 30), 30) as c_ma20_ma30_sd30",]
    # exprs = ["ts_std(ts_mean(ts_mean(close, 20), 30), 30) as c_ma20_ma30_sd30",
    #          "abs(c_ma20_ma30_sd30)",
    #          "volume>0?ts_sum(amount, 30)/ts_sum(volume, 30):null as vwap_30",]
    exprs = OmegaConf.to_container(cfg.exprs)
    # print(exprs)
    start_t = time.time()
    df = db.sql(*exprs, show_progress=True)
    # df = db.data.select(pl.col("close").shift(1).over("asset", order_by=["date", "time"]))
    print(f"Collect elapsed: {(time.time()-start_t):.3f}s")
    print(df)
    # return df

if __name__ == '__main__':
    import os
    # pl.Cojnfig.set_verbose(True)
    # os.environ["POLARS_MAX_THREADS"] = "32"
    # # pl.Config.set_engine_affinity("streaming")
    # # pl.Config.set_engine_affinity("gpu")
    #
    print(f"cpu_num: {os.cpu_count()}")
    print(f"polars.threads: {pl.thread_pool_size()}")
    # # print(f"polars.state: {pl.Config.state()}")
    # res = run(type_=2)
    # print(res)
    run()


# -*- coding: utf-8 -*-
"""
---------------------------------------------
Created on 2025/3/5 21:40
@author: ZhangYundi
@email: yundi.xxii@outlook.com
---------------------------------------------
"""
from __future__ import annotations

import importlib.util
import inspect
import sys
from functools import partial
from pathlib import Path

import logair
import polars as pl
from tqdm.auto import tqdm

from .errors import CompileError, FailError
from .expr import Expr
from .expr import build_dependency_graph, topological_sort

# 动态加载模块
module_name = "udf"
module_path = Path(__file__).parent / "udf" / "__init__.py"
spec = importlib.util.spec_from_file_location(module_name, module_path)
module = importlib.util.module_from_spec(spec)
sys.modules[module_name] = module
spec.loader.exec_module(module)

logger = logair.get_logger(__name__)


class QDF:

    def __init__(self,
                 data: pl.LazyFrame | pl.DataFrame,
                 index: tuple[str] = ("date", "time", "asset"),
                 align: bool = True, ):
        assert isinstance(data, (pl.LazyFrame, pl.DataFrame)), "data must be a polars DataFrame or LazyFrame"
        self.data: pl.LazyFrame = (data
                                   .lazy()
                                   .cast({pl.Decimal: pl.Float64})
                                   .fill_nan(None)
                                   .drop_nulls(subset=index)
                                   .sort(index))
        self.data: pl.DataFrame = self.data.collect()

        self._index: pl.DataFrame = self.data.select(index)
        self.dims = [self._index[name].n_unique() for name in index]

        index_greater_than_one = [index[i] for i, dim in enumerate(self.dims) if dim > 1]
        index_one = [index[i] for i, dim in enumerate(self.dims) if dim == 1]

        if align:
            lev_vals: list[pl.DataFrame] = [self._index.select(name).unique(maintain_order=True) for name in
                                            index_greater_than_one]
            full_index = lev_vals[0]
            for lev_val in lev_vals[1:]:
                full_index = full_index.join(lev_val, how="cross", maintain_order="left")
            self.data = full_index.join(self.data, on=index_greater_than_one, how='left', maintain_order="left")
            if index_one:
                self.data = self.data.with_columns(*[pl.lit(self._index[name][0]).alias(name) for name in index_one])

        self.failed = list()
        self._cols = set(self.data.columns)
        self._expr_cache = dict()  # {expr: pl.col(expr.alias)}
        self._cur_batch_expr_cache = dict()
        # self._alias_expr_map = defaultdict(pl.Expr) # {alias: pl.col(ref:alias)}
        self._fn_map = dict()

    def __str__(self):
        return self.data.__str__()

    def __repr__(self):
        return self.data.__str__()

    def register_udf(self, func: callable, name: str = None):
        name = name if name is not None else func.__name__
        setattr(module, name, func)

    def _to_pl_expr(self, e: Expr) -> pl.Expr:
        alias = e.alias
        if alias in self._cols:
            return pl.col(alias)
        elif e in self._expr_cache:
            return self._expr_cache[e].alias(alias)
        else:
            func = self._fn_map.get(e.fn_name)
            if func is None:
                func = getattr(module, e.fn_name)
                _params = sorted(list(inspect.signature(func).parameters.keys()))
                if "dims" in _params:
                    func = partial(func, dims=self.dims)
                self._fn_map[e.fn_name] = func
            args = list()
            kwargs = dict()
            for arg in e.args:
                if isinstance(arg, Expr):
                    args.append(self._expr_cache[arg])
                    # else:
                    # 拓扑解析依赖结构出错
                elif isinstance(arg, dict):
                    kwargs.update(arg)
                elif isinstance(arg, str):
                    if arg.lower() == "null":
                        args.append(None)
                    else:
                        args.append(pl.col(arg))
                else:
                    args.append(arg)  # or args.append(pl.lit(arg))
            try:
                expr_pl: pl.Expr = func(*args, **kwargs)
                self._cur_batch_expr_cache[e] = pl.col(alias)
                return expr_pl.alias(alias)
            except Exception as error:
                raise CompileError(message=f"{e.fn_name}({', '.join([str(arg) for arg in args])})\n{error}") from error

    def sql(self, *exprs: str, show_progress: bool = False, leave: bool = False) -> pl.DataFrame:
        """
        表达式查询
        Parameters
        ----------
        exprs: str
            表达式，比如 "ts_mean(close, 5) as close_ma5"
        show_progress: bool
            是否展示进度条
        leave: bool
            是否保留进度条
        Returns
        -------
            polars.DataFrame
        """
        self.failed = list()
        exprs_parsed = list()
        for expr in exprs:
            try:
                e_parsed = Expr(expr)
                exprs_parsed.append(e_parsed)
            except Exception as e:
                self.failed.append(FailError(expr, e))
        if self.failed:
            logger.warning(f"sql failed num：{len(self.failed)}/{len(exprs)}: \n {self.failed}")
        graph, indegree, expr_map = build_dependency_graph(exprs_parsed, self._cols)
        lvls: list[list[Expr]] = topological_sort(graph, indegree, expr_map)
        pbar = None
        lvl_num = len(lvls)
        if show_progress:
            pbar = tqdm(total=lvl_num, desc=f"{len(exprs)}", leave=leave)
        for i, batch_exprs in enumerate(lvls):
            if show_progress:
                pbar.set_postfix_str(f"level-{i + 1}:{len(batch_exprs)}")
            batch_exprs = [self._to_pl_expr(e) for e in batch_exprs]
            self.data = self.data.with_columns(*batch_exprs).fill_nan(None)
            self._cols = set(self.data.columns)
            self._expr_cache.update(self._cur_batch_expr_cache)
            if show_progress:
                pbar.update(1)

        final_df = self.data.select(*self._index.columns, *[e.alias for e in exprs_parsed])
        return final_df
